{$APPTYPE CONSOLE}
uses Windows, SysUtils, StrUtils;

const
  DRIVE_NAME = '\\.\PhysicalDrive';
  MAX_DRIVE_COUNT = 15;
  MAX_PARTITION_COUNT = 16;

  FILE_DEVICE_DISK = $00000007;
  FILE_ANY_ACCESS = 0;
  METHOD_BUFFERED = 0;
  IOCTL_DISK_BASE = FILE_DEVICE_DISK;
  IOCTL_DISK_GET_DRIVE_GEOMETRY_EX = (IOCTL_DISK_BASE shl 16) or (FILE_ANY_ACCESS shl 14) or ($0028 shl 2) or (METHOD_BUFFERED);
  IOCTL_DISK_GET_DRIVE_LAYOUT_EX = (IOCTL_DISK_BASE shl 16) or (FILE_ANY_ACCESS shl 14) or ($0014 shl 2) or (METHOD_BUFFERED);
  FILE_DEVICE_MASS_STORAGE = $0000002d;
  IOCTL_STORAGE_BASE = FILE_DEVICE_MASS_STORAGE;
  IOCTL_STORAGE_GET_DEVICE_NUMBER = (IOCTL_STORAGE_BASE shl 16) or (FILE_ANY_ACCESS shl 14) or ($0420 shl 2) or (METHOD_BUFFERED);

type
  TStorageDeviceNumber = record
    DeviceType: DWORD;
    DeviceNumber: DWORD;
    PartitionNumber: DWORD;
  end;

  TMediaType = (
    Unknown,                // Format is unknown
    F5_1Pt2_512,            // 5.25", 1.2MB,  512 bytes/sector
    F3_1Pt44_512,           // 3.5",  1.44MB, 512 bytes/sector
    F3_2Pt88_512,           // 3.5",  2.88MB, 512 bytes/sector
    F3_20Pt8_512,           // 3.5",  20.8MB, 512 bytes/sector
    F3_720_512,             // 3.5",  720KB,  512 bytes/sector
    F5_360_512,             // 5.25", 360KB,  512 bytes/sector
    F5_320_512,             // 5.25", 320KB,  512 bytes/sector
    F5_320_1024,            // 5.25", 320KB,  1024 bytes/sector
    F5_180_512,             // 5.25", 180KB,  512 bytes/sector
    F5_160_512,             // 5.25", 160KB,  512 bytes/sector
    RemovableMedia,         // Removable media other than floppy
    FixedMedia,             // Fixed hard disk media
    F3_120M_512,            // 3.5", 120M Floppy
    F3_640_512,             // 3.5" ,  640KB,  512 bytes/sector
    F5_640_512,             // 5.25",  640KB,  512 bytes/sector
    F5_720_512,             // 5.25",  720KB,  512 bytes/sector
    F3_1Pt2_512,            // 3.5" ,  1.2Mb,  512 bytes/sector
    F3_1Pt23_1024,          // 3.5" ,  1.23Mb, 1024 bytes/sector
    F5_1Pt23_1024,          // 5.25",  1.23MB, 1024 bytes/sector
    F3_128Mb_512,           // 3.5" MO 128Mb   512 bytes/sector
    F3_230Mb_512,           // 3.5" MO 230Mb   512 bytes/sector
    F8_256_128,             // 8",     256KB,  128 bytes/sector
    F3_200Mb_512,           // 3.5",   200M Floppy (HiFD)
    F3_240M_512,            // 3.5",   240Mb Floppy (HiFD)
    F3_32M_512              // 3.5",   32Mb Floppy
  );

  TDiskGeometry = record
    Cylinders: UInt64;
    MediaType: TMediaType;
    TracksPerCylinder: ULONG;
    SectorsPerTrack: ULONG;
    BytesPerSector: ULONG;
  end;

  TPartitionStyle = (
    PARTITION_STYLE_MBR,
    PARTITION_STYLE_GPT,
    PARTITION_STYLE_RAW
  );

  TDiskId = record
    case Byte of
      0: (Signature: ULONG;
          CheckSum: ULONG);
      1: (DiskId: TGUID);
  end;

  TDiskPartitionInfo = record
    SizeOfPartitionInfo: ULONG;
    PartitionStyle: TPartitionStyle;
    DiskId: TDiskId;
  end;

  TDetectionType = (
    DetectNone,
    DetectInt13,
    DetectExInt13);

  TDiskInt13Info = record
    DriveSelect: WORD;
    MaxCylinders: DWORD;
    SectorsPerTrack: WORD;
    MaxHeads: WORD;
    NumberDrives: WORD;
  end;

  TDiskExInt13Info = record
    ExBufferSize: WORD;
    ExFlags: WORD;
    ExCylinders: DWORD;
    ExHeads: DWORD;
    ExSectorsPerTrack: DWORD;
    ExSectorsPerDrive: UInt;
    ExSectorSize: WORD;
    ExReserved: WORD;
  end;

  TDiskDetectionInfo = record
    SizeOfDetectInfo: ULONG;
    DetectionType: TDetectionType;
    Int13: TDiskInt13Info;
    ExInt13: TDiskExInt13Info;
  end;

  TDiskGeometryEx = record
    Geometry: TDiskGeometry;
    DiskSize: UInt64;
    DiskPartitionInfo: TDiskPartitionInfo;
    DiskDetectionInfo: TDiskDetectionInfo;
  end;

  TDiskGeometryEx2 = record
    Geometry: TDiskGeometry;
    DiskSize: UInt64;
    DiskPartitionInfo: record
      SizeOfPartitionInfo: ULONG;
      DiskId: record
        PartitionStyle: record
          Signature: ULONG;
          CheckSum: ULONG;
        end;
      end;
    end;
    DiskDetectionInfo: TDiskDetectionInfo;
  end;
  PDiskGeometryEx2 = ^TDiskGeometryEx2;

  TDriveLayoutInformationMbr = record
    Signature: DWORD;
    CheckSum: DWORD;
  end;

  TDriveLayoutInformationGpt = record
    DiskId: TGUID;
    StartingUsableOffset: UInt64;
    UsableLength: UInt64;
    MaxPartitionCount: DWORD;
  end;

  TDriveLayoutInformationUnion = record
    case Byte of
    0: (Mbr: TDriveLayoutInformationMbr);
    1: (Gpt: TDriveLayoutInformationGpt);
  end;

  TPartitionInformationMbr = record
    PartitionType: Byte;
    BootIndicator: Boolean;
    RecognizedPartition: Boolean;
    HiddenSectors: DWORD;
    PartitionId: TGUID;
  end;

  TPartitionInformationGpt = record
    PartitionType: TGUID;
    PartitionId: TGUID;
    Attributes: UInt64;
    Name: array [0..35] of WideChar;
  end;

  TPartitionInformationEx = record
    PartitionStyle: TPartitionStyle;
    StartingOffset: UInt64;
    PartitionLength: UInt64;
    PartitionNumber: DWORD;
    RewritePartition: Boolean;
    IsServicePartition: Boolean;
    case Byte of
      0: (Mbr: TPartitionInformationMbr);
      1: (Gpt: TPartitionInformationGpt);
  end;

  TDriveLayoutInformationEx = record
    PartitionStyle: TPartitionStyle;
    PartitionCount: DWORD;
    Layout: TDriveLayoutInformationUnion;
    PartitionEntry: array [0..MAX_PARTITION_COUNT-1] of TPartitionInformationEx;
  end;

const
  YesNo: array [Boolean] of String = ('No', 'Yes');
  PartitionStyles: array [TPartitionStyle] of String = ('MBR', 'GPT', 'RAW');
  DetectionTypes: array [TDetectionType] of String = ('None', 'Int13', 'ExInt13');
  MediaTypes: array [TMediaType] of String = ('Unknown',
    'F5_1Pt2_512', 'F3_1Pt44_512', 'F3_2Pt88_512', 'F3_20Pt8_512', 'F3_720_512',
    'F5_360_512', 'F5_320_512', 'F5_320_1024', 'F5_180_512', 'F5_160_512',
    'RemovableMedia', 'FixedMedia', 'F3_120M_512', 'F3_640_512', 'F5_640_512',
    'F5_720_512', 'F3_1Pt2_512', 'F3_1Pt23_1024', 'F5_1Pt23_1024', 'F3_128Mb_512',
    'F3_230Mb_512', 'F8_256_128', 'F3_200Mb_512', 'F3_240M_512', 'F3_32M_512');
  DriveTypes: array [0..6] of String = ('Unknown type', 'Invalid', 'Removable', 'Fixed', 'Remote', 'CD-ROM', 'RAM disk');

function GetVolumePathNamesForVolumeName(lpszVolumeName: LPCSTR; lpszVolumePathNames: LPSTR;
  cchBufferLength: DWORD; var lpcchReturnLength: DWORD): BOOL; stdcall; external kernel32 name 'GetVolumePathNamesForVolumeNameA';

function IntToStrBy3(N: Int64; Sep: Char = ''''): String;
var Len, i: Integer;
begin
  Result := IntToStr(N);
  Len := Length(Result);
  for i := 1 to (Len-1) div 3 do
    Insert(Sep, Result, Len-i*3+1);
end;

function GetDetectionTypeName(N: TDetectionType): String;
begin
  if N <= High(DetectionTypes) then Result := DetectionTypes[N]
  else Result := '<Unknown>';
end;

function GetPartitionStyleName(N: TPartitionStyle): String;
begin
  if N <= High(PartitionStyles) then Result := PartitionStyles[N]
  else Result := '<Unknown>';
end;

function GetMediaTypeName(N: TMediaType): String;
begin
  if N <= High(MediaTypes) then Result := MediaTypes[N]
  else Result := '<Unknown>';
end;

function GetDriveTypeName(N: DWORD): String;
begin
  if N <= High(DriveTypes) then Result := DriveTypes[N]
  else Result := '<Unknown>';
end;

var
  H: Thandle;
  DiskGeometryEx: TDiskGeometryEx;
  DriveLayoutInformationEx: TDriveLayoutInformationEx;
  Volume: String;
  VolName, FileSys: array [0..1023] of Char;
  SerialNumber, MaxCompLen, FileSysFlags: DWORD;
  Error: array [0..255] of Char;
  Returned: DWord;
  D: Char;
  DriveNo, PartNo, k: Integer;
  DeviceNumbers: array ['A'..'Z'] of TStorageDeviceNumber;

begin
  AssignFile(Output, 'HardDriveInfo.txt');
  ReWrite(Output);

  FillChar(DeviceNumbers, SizeOf(DeviceNumbers), -1);
  for D := 'A' to 'Z' do
  begin
    H := CreateFile(PChar('\\.\'+D+':'), 0, 0, nil, OPEN_EXISTING, 0, 0);
    if H <> INVALID_HANDLE_VALUE then
    begin
      DeviceIoControl(H, IOCTL_STORAGE_GET_DEVICE_NUMBER, nil, 0, @DeviceNumbers[D], SizeOf(TStorageDeviceNumber), Returned, nil);
      CloseHandle(H);
    end;
  end;

  for DriveNo := 0 to MAX_DRIVE_COUNT-1 do
  begin
    k := Length(DRIVE_NAME)+2+Integer(DriveNo>=10);
    WriteLn(DupeString('-', k));
    WriteLn(DRIVE_NAME, DriveNo, ':');
    WriteLn(DupeString('-', k));

    H := CreateFile(PChar(DRIVE_NAME+IntToStr(DriveNo)), 0, 0, nil, OPEN_EXISTING, 0, 0);
    if H = INVALID_HANDLE_VALUE then
    begin
      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, GetLastError, LANG_USER_DEFAULT, Error, SizeOf(Error), nil);
      AnsiToOEM(Error, Error);
      Write('>>> Drive open error #', GetLastError, ': ', Error);
      WriteLn;
      Continue;
    end;

    FillChar(DiskGeometryEx, SizeOf(DiskGeometryEx), 0);
    if not DeviceIoControl(H, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, nil, 0, @DiskGeometryEx, SizeOf(DiskGeometryEx), Returned, nil) then
    begin
      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, GetLastError, LANG_USER_DEFAULT, Error, SizeOf(Error), nil);
      AnsiToOEM(Error, Error);
      Write('>>> DeviceIoControl(IOCTL_DISK_GET_DRIVE_GEOMETRY_EX) error #', GetLastError, ': ', Error);
    end
    else with DiskGeometryEx do
    begin
      WriteLn;
      WriteLn('IOCTL_DISK_GET_DRIVE_GEOMETRY_EX:');
      WriteLn('- MediaType: ', GetMediaTypeName(Geometry.MediaType));
      WriteLn('- Cylinders: ', IntToStrBy3(Geometry.Cylinders));
      WriteLn('- TracksPerCylinder: ', IntToStrBy3(Geometry.TracksPerCylinder));
      WriteLn('- SectorsPerTrack: ', IntToStrBy3(Geometry.SectorsPerTrack));
      WriteLn('- BytesPerSector: ', IntToStrBy3(Geometry.BytesPerSector));
      WriteLn('- DiskSize: ', IntToStrBy3(DiskSize));
  
      WriteLn('- PartitionStyle: ', GetPartitionStyleName(DiskPartitionInfo.PartitionStyle));
      case DiskPartitionInfo.SizeOfPartitionInfo of
        24:  // Gpt(?)
        begin
          WriteLn('- GUID: ', GUIDToString(DiskPartitionInfo.DiskId.DiskId));
          AnsiToOEM(Error, Error);
          with DiskDetectionInfo do
            WriteLn('- DetectionType: ', GetDetectionTypeName(DetectionType),
                    ', (Int13.MaxCylinders: ', IntToStrBy3(Int13.MaxCylinders),
                    ', ExInt13.ExCylinders: ', IntToStrBy3(ExInt13.ExCylinders), ')');
        end;
        16:  // Mbr(?)
        begin
          WriteLn('- Signature: ', IntToHex(DiskPartitionInfo.DiskId.Signature, 8), ' (hex)');
          WriteLn('- Checksum: ', IntToHex(DiskPartitionInfo.DiskId.Signature, 8), ' (hex)');
          with PDiskGeometryEx2(@DiskGeometryEx)^.DiskDetectionInfo do
            WriteLn('- DetectionType: ', GetDetectionTypeName(DetectionType),
                    ', (Int13.MaxCylinders: ', IntToStrBy3(Int13.MaxCylinders),
                    ', ExInt13.ExCylinders: ', IntToStrBy3(ExInt13.ExCylinders), ')');
        end;
      end;
    end;

    FillChar(DriveLayoutInformationEx, SizeOf(DriveLayoutInformationEx), 0);
    if not DeviceIoControl(H, IOCTL_DISK_GET_DRIVE_LAYOUT_EX, nil, 0, @DriveLayoutInformationEx, SizeOf(DriveLayoutInformationEx), Returned, nil) then
    begin
      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, GetLastError, LANG_USER_DEFAULT, Error, SizeOf(Error), nil);
      AnsiToOEM(Error, Error);
      Write('>>> DeviceIoControl(IOCTL_DISK_GET_DRIVE_LAYOUT_EX) error #', GetLastError, ': ', Error);
    end
    else with DriveLayoutInformationEx do
    begin
      WriteLn;
      WriteLn('IOCTL_DISK_GET_DRIVE_LAYOUT_EX:');
      WriteLn('- PartitionStyle: ', GetPartitionStyleName(PartitionStyle));
      WriteLn('- PartitionCount: ', IntToStrBy3(PartitionCount));
      case PartitionStyle of
        PARTITION_STYLE_GPT: with Layout.Gpt do
        begin
          WriteLn('- DiskId: ', GUIDToString(DiskId));
          WriteLn('- StartingUsableOffset: ', IntToStrBy3(StartingUsableOffset));
          WriteLn('- UsableLength: ', IntToStrBy3(UsableLength));
          WriteLn('- MaxPartitionCount: ', IntToStrBy3(MaxPartitionCount));
        end;
        PARTITION_STYLE_MBR: with Layout.Mbr do
        begin
          WriteLn('- Signature: ', IntToHex(Signature, 8), ' (hex)');
          WriteLn('- CheckSum: ', IntToHex(CheckSum, 8), ' (hex)');
        end;
      end;

      for PartNo := 0 to PartitionCount-1 do
        with PartitionEntry[PartNo] do
        begin
          if (PartitionStyle = PARTITION_STYLE_MBR) and (not Mbr.RecognizedPartition) then Continue;
          WriteLn;
          WriteLn('> PartitionEntry[', PartNo, ']:');
          WriteLn('  * PartitionStyle: ', GetPartitionStyleName(PartitionStyle));
          WriteLn('  * StartingOffset: ', IntToStrBy3(StartingOffset));
          WriteLn('  * PartitionLength: ', IntToStrBy3(PartitionLength));
          WriteLn('  * PartitionNumber: ', IntToStrBy3(PartitionNumber));
          WriteLn('  * RewritePartition: ', YesNo[RewritePartition]);
          WriteLn('  * IsServicePartition: ', YesNo[IsServicePartition]);
          case PartitionStyle of
            PARTITION_STYLE_GPT: with Gpt do
            begin
              WriteLn('  * PartitionType: ', GUIDToString(PartitionType));
              WriteLn('  * PartitionId: ', GUIDToString(PartitionId));
              WriteLn('  * Attributes: ', IntToHex(Attributes, 16), ' (hex)');
              WriteLn('  * Name: ', Name);
            end;
            PARTITION_STYLE_MBR: with Mbr do
            begin
              WriteLn('  * PartitionType: ', IntToHex(PartitionType, 2), ' (hex)');
              WriteLn('  * BootIndicator: ', YesNo[BootIndicator]);
              WriteLn('  * RecognizedPartition: ', YesNo[RecognizedPartition]);
              WriteLn('  * HiddenSectors: ', IntToStrBy3(HiddenSectors));
              WriteLn('  * PartitionId: ', GUIDToString(PartitionId));
            end;
          end;
          Volume := '';
          for D := 'A' to 'Z' do
            with DeviceNumbers[D] do
              if (DWORD(DriveNo) = DeviceNumber) and (DWORD(PartNo) = PartitionNumber-1) then
              begin
                Volume := D+':';
                Break;
              end;
          if Volume <> '' then
          begin
            WriteLn('  ', Volume);
            WriteLn('    - Drive type: ', GetDriveTypeName(GetDriveType(PChar(Volume))));
            if GetVolumeInformation(PChar(Volume), VolName, SizeOf(VolName), @SerialNumber, MaxCompLen, FileSysFlags, FileSys, SizeOf(FileSys)) then
            begin
              AnsiToOEM(VolName, VolName);
              WriteLn('    - Volume name: ', VolName);
              WriteLn('    - File system: ', FileSys);
              WriteLn('    - Serial number: ', IntToHex(SerialNumber, 8), ' (hex)');
              WriteLn('    - Maximum component length: ', IntToStrBy3(MaxCompLen));
              WriteLn('    - File system flags: ', IntToHex(FileSysFlags, 8), ' (hex)');
            end;
          end;
        end;
    end;
    
    WriteLn;
    CloseHandle(H);
  end;
end.
